@@ -1,21 +1,8 @@
 /*
- * Copyright 2015-2017 Milian Wolff <mail@milianw.de>
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Library General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- */
+    SPDX-FileCopyrightText: 2015-2017 Milian Wolff <mail@milianw.de>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 #ifndef HISTOGRAMMODEL_H
 #define HISTOGRAMMODEL_H
@@ -27,7 +14,8 @@
 struct HistogramColumn
 {
     qint64 allocations;
-    LocationData::Ptr location;
+    qint64 totalAllocated;
+    Symbol symbol;
 };
 Q_DECLARE_TYPEINFO(HistogramColumn, Q_MOVABLE_TYPE);
 
@@ -35,7 +23,7 @@ struct HistogramRow
 {
     HistogramRow()
     {
-        columns.fill({0, {}});
+        columns.fill({0, 0, {}});
     }
     enum
     {
@@ -48,7 +36,12 @@ struct HistogramRow
 Q_DECLARE_TYPEINFO(HistogramRow, Q_MOVABLE_TYPE);
 Q_DECLARE_METATYPE(HistogramRow)
 
-using HistogramData = QVector<HistogramRow>;
+struct HistogramData
+{
+    QVector<HistogramRow> rows;
+    std::shared_ptr<const ResultData> resultData;
+};
+Q_DECLARE_METATYPE(HistogramData)
 
 class HistogramModel : public QAbstractTableModel
 {
