@@ -1,20 +1,8 @@
 /*
- * Copyright 2014-2016 Milian Wolff <mail@milianw.de>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
- */
+    SPDX-FileCopyrightText: 2014-2016 Milian Wolff <mail@milianw.de>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 /**
  * @file heaptrack_print.cpp
@@ -25,11 +13,14 @@
 #include <boost/program_options.hpp>
 
 #include "analyze/accumulatedtracedata.h"
+#include "analyze/suppressions.h"
 
 #include <future>
 #include <iomanip>
 #include <iostream>
 
+#include <tsl/robin_set.h>
+
 #include "util/config.h"
 
 using namespace std;
@@ -51,8 +42,9 @@ struct MergedAllocation : public AllocationData
 class formatBytes
 {
 public:
-    formatBytes(int64_t bytes)
+    formatBytes(int64_t bytes, int width = 0)
         : m_bytes(bytes)
+        , m_width(width)
     {
     }
 
@@ -60,36 +52,45 @@ public:
 
 private:
     int64_t m_bytes;
+    int m_width;
 };
 
+template <typename Bytes>
+ostream& writeBytes(ostream& out, Bytes bytes, int width, const char* unit)
+{
+    const auto unitLength = strlen(unit);
+    if (width > static_cast<int>(unitLength)) {
+        return out << fixed << setprecision(2) << setw(width - unitLength) << bytes << unit;
+    } else {
+        return out << fixed << setprecision(2) << bytes << *unit;
+    }
+}
+
 ostream& operator<<(ostream& out, const formatBytes data)
 {
-    if (data.m_bytes < 0) {
-        // handle negative values
-        return out << '-' << formatBytes(-data.m_bytes);
-    }
-    if (data.m_bytes < 1000) {
-        // no fancy formatting for plain byte values, esp. no .00 factions
-        return out << data.m_bytes << 'B';
-    }
+    auto bytes = static_cast<double>(data.m_bytes);
 
     static const auto units = {"B", "KB", "MB", "GB", "TB"};
     auto unit = units.begin();
     size_t i = 0;
-    double bytes = data.m_bytes;
-    while (i < units.size() - 1 && bytes > 1000.) {
+    while (i < units.size() - 1 && std::abs(bytes) > 1000.) {
         bytes /= 1000.;
         ++i;
         ++unit;
     }
-    return out << fixed << setprecision(2) << bytes << *unit;
+
+    if (i == 0) {
+        // no fractions for bytes
+        return writeBytes(out, data.m_bytes, data.m_width, *unit);
+    } else {
+        return writeBytes(out, bytes, data.m_width, *unit);
+    }
 }
 
 enum CostType
 {
     Allocations,
     Temporary,
-    Allocated,
     Leaked,
     Peak
 };
@@ -102,8 +103,6 @@ std::istream& operator>>(std::istream& in, CostType& type)
         type = Allocations;
     else if (token == "temporary")
         type = Temporary;
-    else if (token == "allocated")
-        type = Allocated;
     else if (token == "leaked")
         type = Leaked;
     else if (token == "peak")
@@ -117,6 +116,7 @@ struct Printer final : public AccumulatedTraceData
 {
     void finalize()
     {
+        applyLeakSuppressions();
         filterAllocations();
         mergedAllocations = mergeAllocations(allocations);
     }
@@ -154,13 +154,10 @@ struct Printer final : public AccumulatedTraceData
         vector<MergedAllocation> ret;
         ret.reserve(allocations.size());
         for (const Allocation& allocation : allocations) {
-            if (allocation.traceIndex) {
-                mergeAllocation(&ret, allocation);
-            }
+            mergeAllocation(&ret, allocation);
         }
         for (MergedAllocation& merged : ret) {
             for (const Allocation& allocation : merged.traces) {
-                merged.allocated += allocation.allocated;
                 merged.allocations += allocation.allocations;
                 merged.leaked += allocation.leaked;
                 merged.peak += allocation.peak;
@@ -278,8 +275,7 @@ struct Printer final : public AccumulatedTraceData
 
     void printBacktrace(TraceNode node, ostream& out, const size_t indent = 0, bool skipFirst = false) const
     {
-        unordered_set<uint32_t> recursionGuard;
-        recursionGuard.insert(node.ipIndex.index);
+        tsl::robin_set<TraceIndex> recursionGuard;
         while (node.ipIndex) {
             const auto& ip = findIp(node.ipIndex);
             if (!skipFirst) {
@@ -291,11 +287,10 @@ struct Printer final : public AccumulatedTraceData
                 break;
             }
 
-            if (!recursionGuard.insert(node.parentIndex.index).second) {
-                cerr << "Trace recursion detected - corrupt data file?" << endl;
+            if (!recursionGuard.insert(node.parentIndex).second) {
+                cerr << "Trace recursion detected - corrupt data file? " << node.parentIndex.index << endl;
                 break;
             }
-
             node = findTrace(node.parentIndex);
         };
     }
@@ -344,6 +339,10 @@ struct Printer final : public AccumulatedTraceData
             label(allocation);
             printIp(allocation.ipIndex, cout);
 
+            if (!allocation.ipIndex) {
+                continue;
+            }
+
             sort(allocation.traces.begin(), allocation.traces.end(), sortOrder);
             int64_t handled = 0;
             for (size_t j = 0; j < min(subPeakLimit, allocation.traces.size()); ++j) {
@@ -516,10 +515,13 @@ struct Printer final : public AccumulatedTraceData
         }
     }
 
-    void handleTimeStamp(int64_t /*oldStamp*/, int64_t newStamp) override
+    void handleTimeStamp(int64_t /*oldStamp*/, int64_t newStamp, bool isFinalTimeStamp, ParsePass pass) override
     {
+        if (pass != ParsePass::FirstPass) {
+            return;
+        }
         if (massifOut.is_open()) {
-            writeMassifSnapshot(newStamp, newStamp == totalTime);
+            writeMassifSnapshot(newStamp, isFinalTimeStamp);
         }
     }
 
@@ -572,8 +574,6 @@ int main(int argc, char** argv)
             "Print backtraces to top allocators, sorted by number of temporary allocations.")
         ("print-leaks,l", po::value<bool>()->default_value(false)->implicit_value(true),
             "Print backtraces to leaked memory allocations.")
-        ("print-overall-allocated,o", po::value<bool>()->default_value(false)->implicit_value(true),
-            "Print top overall allocators, ignoring memory frees.")
         ("peak-limit,n", po::value<size_t>()->default_value(10)->implicit_value(10),
             "Limit the number of reported peaks.")
         ("sub-peak-limit,s", po::value<size_t>()->default_value(5)->implicit_value(5),
@@ -584,7 +584,6 @@ int main(int argc, char** argv)
             "The cost type to use when generating a flamegraph. Possible options are:\n"
             "  - allocations: number of allocations\n"
             "  - temporary: number of temporary allocations\n"
-            "  - allocated: bytes allocated, ignoring deallocations\n"
             "  - leaked: bytes not deallocated at the end\n"
             "  - peak: bytes consumed at highest total memory consumption")
         ("print-flamegraph,F", po::value<string>()->default_value(string()),
@@ -606,6 +605,17 @@ int main(int argc, char** argv)
             "You can set the value to zero to disable detailed snapshots.\n")
         ("filter-bt-function", po::value<string>()->default_value(string()),
             "Only print allocations where the backtrace contains the given function.")
+        ("suppressions", po::value<string>()->default_value(string()),
+            "Load list of leak suppressions from the specified file. Specify one suppression per line, and start each line with 'leak:', i.e. use the LSAN suppression file format.")
+        ("disable-embedded-suppressions",
+            "Ignore suppression definitions that are embedded into the heaptrack data file. By default, heaptrack will copy the suppressions"
+            "optionally defined via a `const char *__lsan_default_suppressions()` symbol in the debuggee application. These are then always "
+            "applied when analyzing the data, unless this feature is explicitly disabled using this command line option.")
+        ("disable-builtin-suppressions",
+            "Ignore suppression definitions that are built into heaptrack. By default, heaptrack will suppress certain "
+            "known leaks from common system libraries.")
+        ("print-suppressions", po::value<bool>()->default_value(false)->implicit_value(true),
+            "Show statistics for matched suppressions.")
         ("help,h", "Show this help message.")
         ("version,v", "Displays version information.");
     // clang-format on
@@ -667,24 +677,33 @@ int main(int argc, char** argv)
         data.massifDetailedFreq = vm["massif-detailed-freq"].as<size_t>();
     }
     const bool printLeaks = vm["print-leaks"].as<bool>();
-    const bool printOverallAlloc = vm["print-overall-allocated"].as<bool>();
     const bool printPeaks = vm["print-peaks"].as<bool>();
     const bool printAllocs = vm["print-allocators"].as<bool>();
     const bool printTemporary = vm["print-temporary"].as<bool>();
+    const auto printSuppressions = vm["print-suppressions"].as<bool>();
+    const auto suppressionsFile = vm["suppressions"].as<string>();
+
+    data.filterParameters.disableEmbeddedSuppressions = vm.count("disable-embedded-suppressions");
+    data.filterParameters.disableBuiltinSuppressions = vm.count("disable-builtin-suppressions");
+    bool suppressionsOk = false;
+    data.filterParameters.suppressions = parseSuppressions(suppressionsFile, &suppressionsOk);
+    if (!suppressionsOk) {
+        return 1;
+    }
 
     cout << "reading file \"" << inputFile << "\" - please wait, this might take some time..." << endl;
 
     if (!diffFile.empty()) {
         cout << "reading diff file \"" << diffFile << "\" - please wait, this might take some time..." << endl;
         Printer diffData;
-        auto diffRead = async(launch::async, [&diffData, diffFile]() { return diffData.read(diffFile); });
+        auto diffRead = async(launch::async, [&diffData, diffFile]() { return diffData.read(diffFile, false); });
 
-        if (!data.read(inputFile) || !diffRead.get()) {
+        if (!data.read(inputFile, false) || !diffRead.get()) {
             return 1;
         }
 
         data.diff(diffData);
-    } else if (!data.read(inputFile)) {
+    } else if (!data.read(inputFile, false)) {
         return 1;
     }
 
@@ -707,20 +726,6 @@ int main(int argc, char** argv)
         cout << endl;
     }
 
-    if (printOverallAlloc) {
-        cout << "MOST BYTES ALLOCATED OVER TIME (ignoring deallocations)\n";
-        data.printAllocations(&AllocationData::allocated,
-                              [](const AllocationData& data) {
-                                  cout << formatBytes(data.allocated) << " allocated over " << data.allocations
-                                       << " calls from\n";
-                              },
-                              [](const AllocationData& data) {
-                                  cout << formatBytes(data.allocated) << " allocated over " << data.allocations
-                                       << " calls from:\n";
-                              });
-        cout << endl;
-    }
-
     if (printPeaks) {
         cout << "PEAK MEMORY CONSUMERS\n";
         data.printAllocations(&AllocationData::peak,
@@ -767,17 +772,31 @@ int main(int argc, char** argv)
         cout << endl;
     }
 
-    const double totalTimeS = 0.001 * data.totalTime;
-    cout << "total runtime: " << fixed << totalTimeS << "s.\n"
-         << "bytes allocated in total (ignoring deallocations): " << formatBytes(data.totalCost.allocated) << " ("
-         << formatBytes(data.totalCost.allocated / totalTimeS) << "/s)" << '\n'
+    const double totalTimeS = data.totalTime ? (1000. / data.totalTime) : 1.;
+    cout << "total runtime: " << fixed << (data.totalTime / 1000.) << "s.\n"
          << "calls to allocation functions: " << data.totalCost.allocations << " ("
-         << int64_t(data.totalCost.allocations / totalTimeS) << "/s)\n"
+         << int64_t(data.totalCost.allocations * totalTimeS) << "/s)\n"
          << "temporary memory allocations: " << data.totalCost.temporary << " ("
-         << int64_t(data.totalCost.temporary / totalTimeS) << "/s)\n"
+         << int64_t(data.totalCost.temporary * totalTimeS) << "/s)\n"
          << "peak heap memory consumption: " << formatBytes(data.totalCost.peak) << '\n'
          << "peak RSS (including heaptrack overhead): " << formatBytes(data.peakRSS * data.systemInfo.pageSize) << '\n'
          << "total memory leaked: " << formatBytes(data.totalCost.leaked) << '\n';
+    if (data.totalLeakedSuppressed) {
+        cout << "suppressed leaks: " << formatBytes(data.totalLeakedSuppressed) << '\n';
+
+        if (printSuppressions) {
+            cout << "Suppressions used:\n";
+            cout << setw(16) << "matches" << ' ' << setw(16) << "leaked"
+                 << " pattern\n";
+            for (const auto& suppression : data.suppressions) {
+                if (!suppression.matches) {
+                    continue;
+                }
+                cout << setw(16) << suppression.matches << ' ' << formatBytes(suppression.leaked, 16) << ' '
+                     << suppression.pattern << '\n';
+            }
+        }
+    }
 
     if (!printHistogram.empty()) {
         ofstream histogram(printHistogram, ios_base::out);
@@ -806,9 +825,6 @@ int main(int argc, char** argv)
                 case Allocations:
                     flamegraph << allocation.allocations;
                     break;
-                case Allocated:
-                    flamegraph << allocation.allocated;
-                    break;
                 case Temporary:
                     flamegraph << allocation.temporary;
                     break;
