@@ -1,20 +1,8 @@
 /*
- * Copyright 2014-2017 Milian Wolff <mail@milianw.de>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
- */
+    SPDX-FileCopyrightText: 2014-2017 Milian Wolff <mail@milianw.de>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 /**
  * @file libheaptrack.cpp
@@ -30,25 +18,41 @@
 #include <link.h>
 #include <pthread.h>
 #include <signal.h>
+#ifdef __linux__
 #include <stdio_ext.h>
-#include <sys/file.h>
 #include <syscall.h>
+#endif
+#ifdef __FreeBSD__
+#include <pthread_np.h>
+#include <sys/types.h>
+#include <sys/sysctl.h>
+#include <sys/user.h>
+#include <libutil.h>
+#endif
+#include <sys/file.h>
 
 #include <atomic>
 #include <cinttypes>
 #include <memory>
 #include <mutex>
+#include <sstream>
 #include <string>
 #include <thread>
-#include <unordered_set>
-
-#include <boost/algorithm/string/replace.hpp>
 
 #include "tracetree.h"
 #include "util/config.h"
 #include "util/libunwind_config.h"
 #include "util/linewriter.h"
 
+extern "C" {
+// see upstream "documentation" at:
+// https://github.com/llvm-mirror/compiler-rt/blob/master/include/sanitizer/lsan_interface.h#L76
+__attribute__((weak)) const char* __lsan_default_suppressions();
+}
+namespace __gnu_cxx {
+__attribute__((weak)) extern void __freeres();
+}
+
 /**
  * uncomment this to get extended debug code for known pointers
  * there are still some malloc functions I'm missing apparently,
@@ -56,6 +60,10 @@
  */
 // #define DEBUG_MALLOC_PTRS
 
+#ifdef DEBUG_MALLOC_PTRS
+#include <tsl/robin_set.h>
+#endif
+
 using namespace std;
 
 namespace {
@@ -74,7 +82,11 @@ chrono::milliseconds elapsedTime()
 
 __pid_t gettid()
 {
+#ifdef __linux__
     return syscall(SYS_gettid);
+#elif defined(__FreeBSD__)
+    return pthread_getthreadid_np();
+#endif
 }
 
 /**
@@ -102,6 +114,7 @@ thread_local bool RecursionGuard::isActive = false;
 
 enum DebugVerbosity
 {
+    WarningOutput,
     NoDebugOutput,
     MinimalOutput,
     VerboseOutput,
@@ -115,51 +128,42 @@ constexpr const DebugVerbosity s_debugVerbosity = NoDebugOutput;
  * Call this to optionally show debug information but give the compiler
  * a hand in removing it all if debug output is disabled.
  */
-template <DebugVerbosity debugLevel, typename... Args>
-inline void debugLog(const char fmt[], Args... args)
+template <DebugVerbosity debugLevel, typename Callback>
+inline void debugLog(Callback callback)
 {
     if (debugLevel <= s_debugVerbosity) {
         RecursionGuard guard;
         flockfile(stderr);
-        fprintf(stderr, "heaptrack debug(%d) [%d:%d]@%lu ", static_cast<int>(debugLevel), getpid(), gettid(),
-                elapsedTime().count());
-        fprintf(stderr, fmt, args...);
+        if (debugLevel == WarningOutput) {
+            fprintf(stderr, "heaptrack warning [%d:%d]@%" PRIu64 " ", getpid(), gettid(), elapsedTime().count());
+        } else {
+            fprintf(stderr, "heaptrack debug(%d) [%d:%d]@%" PRIu64 " ", debugLevel, getpid(), gettid(),
+                    elapsedTime().count());
+        }
+        callback(stderr);
         fputc('\n', stderr);
         funlockfile(stderr);
     }
 }
 
+/**
+ * Call this to optionally show debug information but give the compiler
+ * a hand in removing it all if debug output is disabled.
+ */
+template <DebugVerbosity debugLevel, typename... Args>
+inline void debugLog(const char fmt[], Args... args)
+{
+    debugLog<debugLevel>([&](FILE* out) { fprintf(out, fmt, args...); });
+}
+
 void printBacktrace()
 {
     if (s_debugVerbosity == NoDebugOutput)
         return;
 
-#if LIBUNWIND_HAS_UNW_GETCONTEXT && LIBUNWIND_HAS_UNW_INIT_LOCAL
     RecursionGuard guard;
 
-    unw_context_t context;
-    unw_getcontext(&context);
-
-    unw_cursor_t cursor;
-    unw_init_local(&cursor, &context);
-
-    int frameNr = 0;
-    while (unw_step(&cursor)) {
-        ++frameNr;
-        unw_word_t ip = 0;
-        unw_get_reg(&cursor, UNW_REG_IP, &ip);
-
-        unw_word_t sp = 0;
-        unw_get_reg(&cursor, UNW_REG_SP, &sp);
-
-        char symbol[256] = {"<unknown>"};
-        unw_word_t offset = 0;
-        unw_get_proc_name(&cursor, symbol, sizeof(symbol), &offset);
-
-        fprintf(stderr, "#%-2d 0x%016" PRIxPTR " sp=0x%016" PRIxPTR " %s + 0x%" PRIxPTR "\n", frameNr,
-                static_cast<uintptr_t>(ip), static_cast<uintptr_t>(sp), symbol, static_cast<uintptr_t>(offset));
-    }
-#endif
+    Trace::print();
 }
 
 /**
@@ -174,6 +178,27 @@ atomic<bool> s_atexit{false};
  */
 atomic<bool> s_forceCleanup{false};
 
+// based on: https://stackoverflow.com/a/24315631/35250
+void replaceAll(string& str, const string& search, const string& replace)
+{
+    size_t start_pos = 0;
+    while ((start_pos = str.find(search, start_pos)) != string::npos) {
+        str.replace(start_pos, search.length(), replace);
+        start_pos += replace.length();
+    }
+}
+
+// see https://bugs.kde.org/show_bug.cgi?id=408547
+// apparently sometimes flock can return EAGAIN, despite that not being a documented return value
+static int lockFile(int fd)
+{
+    int ret = -1;
+    while ((ret = flock(fd, LOCK_EX | LOCK_NB)) == EAGAIN) {
+        // try again
+    }
+    return ret;
+}
+
 int createFile(const char* fileName)
 {
     string outputFileName;
@@ -194,7 +219,7 @@ int createFile(const char* fileName)
         outputFileName = "heaptrack.$$";
     }
 
-    boost::replace_all(outputFileName, "$$", to_string(getpid()));
+    replaceAll(outputFileName, "$$", to_string(getpid()));
 
     auto out = open(outputFileName.c_str(), O_CREAT | O_WRONLY | O_CLOEXEC, 0644);
     debugLog<VerboseOutput>("will write to %s/%p\n", outputFileName.c_str(), out);
@@ -202,7 +227,17 @@ int createFile(const char* fileName)
     if (out == -1) {
         fprintf(stderr, "ERROR: failed to open heaptrack output file %s: %s (%d)\n", outputFileName.c_str(),
                 strerror(errno), errno);
-    } else if (flock(out, LOCK_EX | LOCK_NB) != 0) {
+    } else if (lockFile(out) != 0) {
+#ifdef __FreeBSD__
+        // pipes do not support flock, create a regular file
+        auto lockpath = outputFileName + ".lock";
+        auto lockfile = open(lockpath.c_str(), O_CREAT | O_WRONLY | O_CLOEXEC, 0644);
+        debugLog<VerboseOutput>("will lock %s/%p\n", lockpath.c_str(), lockfile);
+        if (lockFile(lockfile) == 0) {
+            // leaking the fd seems fine
+            return out;
+        }
+#endif
         fprintf(stderr, "ERROR: failed to lock heaptrack output file %s: %s (%d)\n", outputFileName.c_str(),
                 strerror(errno), errno);
         close(out);
@@ -216,8 +251,7 @@ int createFile(const char* fileName)
  * Thread-Safe heaptrack API
  *
  * The only critical section in libheaptrack is the output of the data,
- * dl_iterate_phdr
- * calls, as well as initialization and shutdown.
+ * dl_iterate_phdr calls, as well as initialization and shutdown.
  */
 class HeapTrack
 {
@@ -254,15 +288,8 @@ public:
         static once_flag once;
         call_once(once, [] {
             debugLog<MinimalOutput>("%s", "doing once-only initialization");
-            // configure libunwind for better speed
-            if (unw_set_caching_policy(unw_local_addr_space, UNW_CACHE_PER_THREAD)) {
-                fprintf(stderr, "WARNING: Failed to enable per-thread libunwind caching.\n");
-            }
-#if LIBUNWIND_HAS_UNW_SET_CACHE_SIZE
-            if (unw_set_cache_size(unw_local_addr_space, 1024, 0)) {
-                fprintf(stderr, "WARNING: Failed to set libunwind cache size.\n");
-            }
-#endif
+
+            Trace::setup();
 
             // do not trace forked child processes
             // TODO: make this configurable
@@ -273,6 +300,13 @@ public:
                     return;
                 }
                 debugLog<MinimalOutput>("%s", "atexit()");
+
+                // free internal libstdc++ resources
+                // see also Valgrind's `--run-cxx-freeres` option
+                if (&__gnu_cxx::__freeres) {
+                    __gnu_cxx::__freeres();
+                }
+
                 s_atexit.store(true);
                 heaptrack_stop();
             });
@@ -293,6 +327,7 @@ public:
         writeExe();
         writeCommandLine();
         writeSystemInfo();
+        writeSuppressions();
 
         if (initAfterCallback) {
             debugLog<MinimalOutput>("%s", "calling initAfterCallback");
@@ -350,19 +385,45 @@ public:
 
     void writeRSS()
     {
-        if (!s_data || !s_data->out.canWrite() || !s_data->procStatm) {
+        if (!s_data || !s_data->out.canWrite()) {
             return;
         }
 
-        // read RSS in pages from statm, then rewind for next read
         size_t rss = 0;
-        if (fscanf(s_data->procStatm, "%*x %zx", &rss) != 1) {
-            fprintf(stderr, "WARNING: Failed to read RSS value from /proc/self/statm.\n");
-            fclose(s_data->procStatm);
-            s_data->procStatm = nullptr;
+
+#ifdef __linux__
+        if (s_data->procStatm == -1) {
             return;
         }
-        rewind(s_data->procStatm);
+        // read RSS in pages from statm, then rewind for next read
+        // NOTE: don't use fscanf here, it could potentially deadlock us
+        const int BUF_SIZE = 512;
+        char buf[BUF_SIZE + 1];
+        if (read(s_data->procStatm, buf, BUF_SIZE) <= 0) {
+            fprintf(stderr, "WARNING: Failed to read RSS value from /proc/self/statm.\n");
+            close(s_data->procStatm);
+            s_data->procStatm = -1;
+            return;
+        }
+        lseek(s_data->procStatm, 0, SEEK_SET);
+
+        if (sscanf(buf, "%*u %zu", &rss) != 1) {
+            fprintf(stderr, "WARNING: Failed to read RSS value from /proc/self/statm.\n");
+            close(s_data->procStatm);
+            s_data->procStatm = -1;
+            return;
+        }
+#elif defined(__FreeBSD__)
+        auto proc_info = kinfo_getproc(getpid());
+        if (proc_info == nullptr) {
+            return;
+        }
+
+        rss = proc_info->ki_rssize;
+
+        free(proc_info);
+#endif
+
         // TODO: compare to rusage.ru_maxrss (getrusage) to find "real" peak?
         // TODO: use custom allocators with known page sizes to prevent tainting
         //       the RSS numbers with heaptrack-internal data
@@ -380,10 +441,18 @@ public:
     {
         const int BUF_SIZE = 1023;
         char buf[BUF_SIZE + 1];
+
+#ifdef __linux__
         ssize_t size = readlink("/proc/self/exe", buf, BUF_SIZE);
+#elif defined(__FreeBSD__)
+        int mib[4] = { CTL_KERN, KERN_PROC, KERN_PROC_PATHNAME, -1 };
+        size_t size = BUF_SIZE;
+        sysctl(mib, 4, buf, &size, NULL, 0);
+#endif
+
         if (size > 0 && size < BUF_SIZE) {
             buf[size] = 0;
-            s_data->out.write("x %s\n", buf);
+            s_data->out.write("x %x %s\n", size, buf);
         }
     }
 
@@ -391,9 +460,18 @@ public:
     {
         s_data->out.write("X");
         const int BUF_SIZE = 4096;
-        char buf[BUF_SIZE + 1];
+        char buf[BUF_SIZE + 1] = {0};
+
+#ifdef __linux__
         auto fd = open("/proc/self/cmdline", O_RDONLY);
         int bytesRead = read(fd, buf, BUF_SIZE);
+        close(fd);
+#elif defined(__FreeBSD__)
+        int mib[4] = { CTL_KERN, KERN_PROC, KERN_PROC_ARGS, getpid() };
+        size_t bytesRead = BUF_SIZE;
+        sysctl(mib, 4, buf, &bytesRead, NULL, 0);
+#endif
+
         char* end = buf + bytesRead;
         for (char* p = buf; p < end;) {
             s_data->out.write(" %s", p);
@@ -401,7 +479,6 @@ public:
                 ; // skip until start of next 0-terminated section
         }
 
-        close(fd);
         s_data->out.write("\n");
     }
 
@@ -411,6 +488,24 @@ public:
                                  static_cast<size_t>(sysconf(_SC_PHYS_PAGES)));
     }
 
+    void writeSuppressions()
+    {
+        if (!__lsan_default_suppressions)
+            return;
+
+        const char* suppressions = __lsan_default_suppressions();
+        if (!suppressions)
+            return;
+
+        std::istringstream stream(suppressions);
+        std::string line;
+        while (std::getline(stream, line)) {
+            s_data->out.write("S ");
+            s_data->out.write(line);
+            s_data->out.write("\n");
+        }
+    }
+
     void handleMalloc(void* ptr, size_t size, const Trace& trace)
     {
         if (!s_data || !s_data->out.canWrite()) {
@@ -418,8 +513,15 @@ public:
         }
         updateModuleCache();
 
-        const auto index = s_data->traceTree.index(
-            trace, [this](uintptr_t ip, uint32_t index) { return s_data->out.writeHexLine('t', ip, index); });
+        const auto index = s_data->traceTree.index(trace, [](uintptr_t ip, uint32_t index) {
+            // decrement addresses by one - otherwise we misattribute the cost to the wrong instruction
+            // for some reason, it seems like we always get the instruction _after_ the one we are interested in
+            // see also: https://github.com/libunwind/libunwind/issues/287
+            // and https://bugs.kde.org/show_bug.cgi?id=439897
+            --ip;
+
+            return s_data->out.writeHexLine('t', ip, index);
+        });
 
 #ifdef DEBUG_MALLOC_PTRS
         auto it = s_data->known.find(ptr);
@@ -445,6 +547,16 @@ public:
         s_data->out.writeHexLine('-', reinterpret_cast<uintptr_t>(ptr));
     }
 
+    static bool isPaused()
+    {
+        return s_paused;
+    }
+
+    static void setPaused(bool state)
+    {
+        s_paused = state;
+    }
+
 private:
     static int dl_iterate_phdr_callback(struct dl_phdr_info* info, size_t /*size*/, void* data)
     {
@@ -456,7 +568,7 @@ private:
 
         debugLog<VerboseOutput>("dlopen_notify_callback: %s %zx", fileName, info->dlpi_addr);
 
-        if (!heaptrack->s_data->out.write("m %s %zx", fileName, info->dlpi_addr)) {
+        if (!heaptrack->s_data->out.write("m %x %s %zx", strlen(fileName), fileName, info->dlpi_addr)) {
             return 1;
         }
 
@@ -505,7 +617,7 @@ private:
             return;
         }
         debugLog<MinimalOutput>("%s", "updateModuleCache()");
-        if (!s_data->out.write("m -\n")) {
+        if (!s_data->out.write("m 1 -\n")) {
             return;
         }
         dl_iterate_phdr(&dl_iterate_phdr_callback, this);
@@ -545,14 +657,12 @@ private:
         {
 
             debugLog<MinimalOutput>("%s", "constructing LockedData");
-            procStatm = fopen("/proc/self/statm", "r");
-            if (!procStatm) {
+#ifdef __linux__
+            procStatm = open("/proc/self/statm", O_RDONLY);
+            if (procStatm == -1) {
                 fprintf(stderr, "WARNING: Failed to open /proc/self/statm for reading: %s.\n", strerror(errno));
-            } else if (setvbuf(procStatm, nullptr, _IONBF, 0)) {
-                // disable buffering to ensure we read the latest values
-                fprintf(stderr, "WARNING: Failed to disable buffering for reading of /proc/self/statm: %s.\n",
-                        strerror(errno));
             }
+#endif
 
             // ensure this utility thread is not handling any signals
             // our host application may assume only one specific thread
@@ -568,10 +678,22 @@ private:
             }
 
             // the mask we set above will be inherited by the thread that we spawn below
-            timerThread = thread([&]() {
+            timerThread = std::thread([&]() {
                 RecursionGuard::isActive = true;
                 debugLog<MinimalOutput>("%s", "timer thread started");
 
+                // HACK: throw the exception once and directly catch it
+                //       without this, tst_inject reproducibly calls
+                //       std::terminate instead of catching the exception
+                //       in the loop below
+                //       I suspect it's some strange side-effect of heaptrack
+                //       intercepting the memory allocations that happen when
+                //       an exception is thrown?
+                try {
+                    throw LockCheckFailed {};
+                } catch (LockCheckFailed) {
+                }
+
                 // now loop and repeatedly print the timestamp and RSS usage to the data stream
                 while (!stopTimerThread) {
                     // TODO: make interval customizable
@@ -600,14 +722,14 @@ private:
             if (timerThread.joinable()) {
                 try {
                     timerThread.join();
-                } catch (std::system_error) {
+                } catch (const std::system_error&) {
                 }
             }
 
             out.close();
 
-            if (procStatm) {
-                fclose(procStatm);
+            if (procStatm != -1) {
+                close(procStatm);
             }
 
             if (stopCallback && (!s_atexit || s_forceCleanup)) {
@@ -618,8 +740,8 @@ private:
 
         LineWriter out;
 
-        /// /proc/self/statm file stream to read RSS value from
-        FILE* procStatm = nullptr;
+        /// /proc/self/statm file descriptor to read RSS value from
+        int procStatm = -1;
 
         /**
          * Calls to dlopen/dlclose mark the cache as dirty.
@@ -633,22 +755,27 @@ private:
         TraceTree traceTree;
 
         atomic<bool> stopTimerThread{false};
-        thread timerThread;
+        std::thread timerThread;
 
         heaptrack_callback_t stopCallback = nullptr;
 
 #ifdef DEBUG_MALLOC_PTRS
-        unordered_set<void*> known;
+        tsl::robin_set<void*> known;
 #endif
     };
 
     static std::mutex s_lock;
     static LockedData* s_data;
+
+private:
+    static std::atomic<bool> s_paused;
 };
 
 std::mutex HeapTrack::s_lock;
 HeapTrack::LockedData* HeapTrack::s_data{nullptr};
+std::atomic<bool> HeapTrack::s_paused{false};
 }
+
 extern "C" {
 
 void heaptrack_init(const char* outputFileName, heaptrack_callback_t initBeforeCallback,
@@ -679,15 +806,25 @@ void heaptrack_stop()
     heaptrack.shutdown();
 }
 
+void heaptrack_pause()
+{
+    HeapTrack::setPaused(true);
+}
+
+void heaptrack_resume()
+{
+    HeapTrack::setPaused(false);
+}
+
 void heaptrack_malloc(void* ptr, size_t size)
 {
-    if (ptr && !RecursionGuard::isActive) {
+    if (!HeapTrack::isPaused() && ptr && !RecursionGuard::isActive) {
         RecursionGuard guard;
 
         debugLog<VeryVerboseOutput>("heaptrack_malloc(%p, %zu)", ptr, size);
 
         Trace trace;
-        trace.fill(2 + HEAPTRACK_DEBUG_BUILD);
+        trace.fill(2 + HEAPTRACK_DEBUG_BUILD * 2);
 
         HeapTrack heaptrack(guard);
         heaptrack.handleMalloc(ptr, size, trace);
@@ -696,7 +833,7 @@ void heaptrack_malloc(void* ptr, size_t size)
 
 void heaptrack_free(void* ptr)
 {
-    if (ptr && !RecursionGuard::isActive) {
+    if (!HeapTrack::isPaused() && ptr && !RecursionGuard::isActive) {
         RecursionGuard guard;
 
         debugLog<VeryVerboseOutput>("heaptrack_free(%p)", ptr);
@@ -708,13 +845,13 @@ void heaptrack_free(void* ptr)
 
 void heaptrack_realloc(void* ptr_in, size_t size, void* ptr_out)
 {
-    if (ptr_out && !RecursionGuard::isActive) {
+    if (!HeapTrack::isPaused() && ptr_out && !RecursionGuard::isActive) {
         RecursionGuard guard;
 
         debugLog<VeryVerboseOutput>("heaptrack_realloc(%p, %zu, %p)", ptr_in, size, ptr_out);
 
         Trace trace;
-        trace.fill(2 + HEAPTRACK_DEBUG_BUILD);
+        trace.fill(2 + HEAPTRACK_DEBUG_BUILD * 2);
 
         HeapTrack heaptrack(guard);
         if (ptr_in) {
@@ -733,4 +870,11 @@ void heaptrack_invalidate_module_cache()
     HeapTrack heaptrack(guard);
     heaptrack.invalidateModuleCache();
 }
+
+void heaptrack_warning(heaptrack_warning_callback_t callback)
+{
+    RecursionGuard guard;
+
+    debugLog<WarningOutput>(callback);
+}
 }
