@@ -1,31 +1,18 @@
 /*
- * Copyright 2014-2017 Milian Wolff <mail@milianw.de>
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU Library General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc.,
- * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- */
+    SPDX-FileCopyrightText: 2014-2017 Milian Wolff <mail@milianw.de>
+
+    SPDX-License-Identifier: LGPL-2.1-or-later
+*/
 
 #include "3rdparty/catch.hpp"
-#include "src/track/trace.h"
+#include "track/trace.h"
 
 #include <algorithm>
 
 using namespace std;
 
 namespace {
-bool fill(Trace& trace, int depth, int skip)
+bool __attribute__((noinline)) fill(Trace& trace, int depth, int skip)
 {
     if (!depth) {
         return trace.fill(skip);
